
#define TABLE_MAX (64)
#define TABLE_TYPE_MAX (16)

struct entry {
	lock_t lock;
	void *element;
};

struct binary_entry {
	char balance;
	struct entry entry;
	struct binary_entry *left, *right;
};

union binary_node {
	struct binary_entry entry;
	union binary_node *next;
};

struct list_entry {
	struct entry entry;
	struct list_entry *previous, *next;
};

union list_node {
	struct list_entry entry;
	union list_node *next;
};

struct list_base {
	struct list_entry **head;
	union list_node *pool;
};

union hash_base {
	void *closed;
	struct list_entry *open;
};

union table_base {
	union binary_node *binary;
	union hash_base hash;
	struct list_base list;
};

enum table_process { USER_TABLE, KERNEL_TABLE };
enum table_duplicates { NO_DUPLICATES, DUPLICATES = 32 };
enum table_type { LIST_TABLE = 2, TREE_TABLE = 0, HASH_TABLE = 16 };
enum tree_table_type { BINARY_TREE, TWO_THREE_TREE = 10 };
enum tree_table_thread_type { UNTHREADED_TREE, THREADED_TREE = 64 };
enum binary_table_type { PLAIN_TREE = 4, BALANCED_TREE = AVL_TREE = 6, PROPER_TREE = 8 };
enum hash_table_type { CLOSED_HASH = 12, OPEN_HASH = 14 };
enum open_hash_table_type { UNORDERED_HASH, ORDERED_HASH = 64, STATIC_HASH, DYNAMIC_HASH = 128 };

typedef int (*compare_fn)(void *, void *);
typedef int (*hash_fn)(void *, int);

struct table {
	int type;
	int size[2];
	int count[2];
	union table_base base;
	union table_avail avail;
	union table_ptr ptr;
	compare_fn compare;
	hash_fn hash;
};

struct table_op {
	struct table *(*create)(struct table *table, int size);
	void (*destroy)(struct table *table);
	void *(*lookup)(struct table *table, void *element);
	struct table *(*insert)(struct table *table, void *element);
	struct table *(*delete)(struct table *table, void *element);
	struct table *(*resize)(struct table *table, int size);
};

union table_node {
	struct table table;
	union table_node *next;
};

struct table_pool {
	union table_node base[TABLE_MAX], *avail;
	int size;
	pthread_mutex_t free;
};

struct table_pool table_pool;
struct table_op *table_op[TABLE_TYPE_MAX]; /* This is common to all process's */

typedef const int table_t;

table_t table_create(int type, int size, compare_fn compare, hash_fn hash) {
	table_t new;

	new = table_op[type];
}
