
#include"visit.h"

static lock_t visit_create;

visit_t visit_create() {
	static struct visit_table *table;
	struct visit_table *new;

	if(lock(&visit_create) > 0) {
		if(table = malloc(VISIT_TABLE_SIZE)) {
			new = table;
			unlock(&visit_create);
			lock_init(&new->pool.visit.free);
			lock_init(&new->pool.thread.free);
			new->pool.visit.avail = next = new->pool.visit.base;
			new->pool.thread.avail = after = new->pool.thread.base;
			for(pos = 0; pos < VISIT_NODE_MAX - 1; pos++) {
				new->pool.visit.base[pos].next = (++next);
				new->pool.thread.base[pos].next = (++after);
			}
			new->pool.visit.base[pos].visit.node = NULL;
			new->pool.visit.base[VISIT_NODE_MAX - 1].next = NULL;
			new->pool.thread.base[VISIT_NODE_MAX - 1].next = NULL;
			new->root = &new->base[VISIT_NODE_MAX - 1];
			new->root->visit = NULL;
			return (visit_t)new;
		}
		unlock(&visit_create);
	}
	return NULL;
}

void visit_destroy(visit_t visit) {
	free(visit);
	return;
}

static struct visit *lookup(visit_t visit, struct visit_entry *new) {
	struct visit_entry *entry, *parent, *child;
	struct visit *found, *initial;
	unsigned key, weight;
	bool_t insert;

	insert = new >= visit->base && new < visit->base + VISIT_MAX;
	key = weight = (UINT_MAX / 2) + 1;
	parent = former = found = NULL;
	initial = new->visit;
	entry = visit->root;
	do {
		weight /= 2;
		lock(&entry->lock);
		if(entry->vist->node < new->visit->node) {
			if((unsigned)new->visit->node < key) {
				if(insert) {
					visit *temp;

					temp = entry->visit;
					entry->visit = new->visit;
					new->visit = temp;
				}
				else break;
			}
			else {
				child = entry->right;
			}
			key += weight;
		}
		else {
			if(entry->vist->node > new->visit->node) {
				if((unsigned)new->visit->node > key) {
					if(insert) {
						visit *temp;

						temp = entry->visit;
						entry->visit = new->visit;
						new->visit = temp;
					}
					else break;
				}
				else {
					child = entry->left;
				}
				key -= weight;
			}
			else {
				found = entry->visit;
				break;
			}
		}
		if(parent) unlock(&parent->lock);
		parent = entry;
		entry = child;
	} while(entry);
	if(insert) {
		if(found) {
			found->queue.rear = found->queue.rear->next = new->queue.front;
		}
		else {
			found = initial;
			if(new->visit->node) {
				lock_init(&new->lock);
				if(parent->visit->node < new->visit->node) parent->left = new;
				else parent->right = new;
			}
		}
	}
	else {
		if(found) {
			if(found->queue.head->thread == pthread_self()) {
				if(! found->queue.head->next) {
					struct visit_entry *successor;

					successor = NULL;
					child = entry->right;
					while(child) {
						if(successor) unlock(&successor->lock);
						lock(&child->lock);
						successor = child;
						child = child->left;
					}
					if(parent) {
						if(parent->left == entry) {
							parent->left = successor;
						}
						else {
							parent->right = successor;
						}
					}
					else {
						if(successor) visit->root = successor;
					}
					if(successor) unlock(&successor->lock);
					found->node = NULL;
				}
			}
			else found = NULL;
		}
	}
	if(parent) unlock(&parent->lock);
	if(entry) unlock(&entry->lock);
	return found;
}

void visit(visit_t visit, void *node) {
	struct visit_entry *new, *found;
	struct visit_thread *thread;
	union visit_node *request;
	pthread_t self;

	while(1) {
		if(thread = visit->pool.thread.avail) {
			if(trylock(&visit->pool.thread.free) > 0) {
				visit->pool.thread.avail = thread->next;
				unlock(&visit->pool.thread.free);
				break;
			}
		}
		pthread_switch();
	}
	thread->thread = self = pthread_self();
	thread->next = NULL;
	while(1) {
		if(request = visit->pool.visit.avail) {
			if(trylock(&visit->pool.visit.free) > 0) {
				visit->pool.visit.avail = request->next;
				unlock(&visit->pool.visit.free);
				break;
			}
		}
		pthread_switch();
	}
	request->visit->node = node;
	request->visit->queue.front = request->visit->queue.rear = thread;
	new = &visit->base[request - visit->pool.visit.base];
	new->visit = &request->visit;
	lock_init(&new->lock);
	if((found = lookup(new)) != &request->visit) {
		lock(&visit->pool.visit.free);
		request->next = visit->pool.visit.avail;
		visit->pool.visit.avail = request;
		unlock(&visit->pool.visit.free);
	}
	while(found->queue.front->thread != self) pthread_switch();
	return;
}

void farewell(visit_t visit, void *node) {
	struct visit request;
	union visit_node *found;
	struct visit_entry old;

	request.node = node;
	lock_init(&old.lock);
	old.visit = &request;
	if(found = (union visit_node *)lookup(old)) {
		struct visit_thread *head;

		head = found->visit.queue.head;
		found->visit.queue.head = head->next;
		lock(&visit->pool.thread.free);
		head->next = visit->pool.thread.avail;
		visit->pool.thread.avail = head;
		unlock(&visit->pool.thread.free);
		if(! found->visit.queue.head && visit->root->visit != found) {
			lock(&visit->pool.visit.free);
			found->next = visit->pool.visit.avail;
			visit->pool.visit.avail = found;
			unlock(&visit->pool.visit.free);
		}
	}
	return;
}
