#ifndef LIST_H
#define LIST_H

#define LIST_MAX (100)
#define LIST_TYPE_MAX (4)

struct single_node {
	struct entry entry;
	struct single_node *next;
};

struct double_node {
	struct entry entry;
	struct double_node *previous, *next;
};

union linked_node {
	struct single_node *single;
	struct double_node *double_;
};

struct linked_base {
	union linked_node base, avail;
};

union list_base {
	void *array;
	struct linked_base linked;
};

union list_end {
	struct array_end {
		atomic_t first, last;
	} array;
	struct list_end {
		union linked_node first, last;
	} list;
};

enum list_insert_type { INSERT_FIRST = 1, INSERT_BETW = 2, INSERT_LAST = 4 };
enum list_delete_type { DELETE_FIRST = 8, DELETE_BETW = 16, DELETE_LAST = 32 };
enum linked_type { DOUBLY_LINKED, SINGLY_LINKED = 64 };

#define STACK_TYPE (9)
#define QUEUE_TYPE (12)
#define LINKED_TYPE (63)

#define LIST_LINKED (INSERT_BETW | DELETE_BETW)

typedef union list_ptr {
	atomic_t array;
	union linked_node linked;
} list_ptr_t;

struct list {
	atomic_t type;
	union list_base base;
	union list_end end;
	atomic_t size;
	atomic_t count;
	lock_t free, ptr;
	counter_t counter;
	struct lock_pool pool;
};

union list_node {
	struct list list;
	union list_node *next;
};

struct list_pool {
	union list_node base[LIST_MAX], *avail;
	lock_t free;
};

typedef const struct list * const list_t;

list_t list_create(int type, int size);
void list_destroy(list_t list);

list_ptr_t *list_first(list_t list, list_ptr_t *ptr);
list_ptr_t *list_last(list_t list, list_ptr_t *ptr);

list_t list_insert(list_t list, void *element, list_ptr_t *ptr);
list_t list_delete(list_t list, list_ptr_t *ptr);

#define top(list) ()
#define push(list,element) (list_insert(list, element, NULL))
#define pop(list) (list_delete(list, NULL))

#define front(list)
#define rear(list)
#define join(list,element) (list_append(list, element, NULL))
#define join_front(list,element,ptr) (list_insert(list, element, NULL))
#define leave(list) (list_delete(list, NULL))
list_t leave_rear(list_t list);

void *list_get(list_t list, int get, list_ptr_t *ptr);

list_t *list_resize(list_t *list, int size);

#endif
