#ifndef TABLE_H
#define TABLE_H

#include"lock.h"
#include"counter.h"

/* Maximum number of tables per process */
#define TABLE_MAX (64)

/* Maximum number of distinct table types */
#define TABLE_TYPE_MAX (16)

enum table_duplicates { NO_DUPLICATES, DUPLICATES = 8 };
enum table_type { LIST_TABLE = 16, TREE_TABLE = 0, HASH_TABLE = 128 };

struct table_base {
	void *data, *avail;
	union {
		struct binary_base binary;
		struct hash_base hash;
		struct list_base list;
	} table;
};

typedef int (*compare_fn)(void *, void *);

union table_parameter {
	struct balance {
		int sum, ratio;
	} balance;
	struct hash {
		float occupancy;
		union  {
			hash_fn hash, rehash;
		} fn;
	} hash;
};

struct table {
	int type;
	int size[3];
	int count[3];
	union table_base base;
	compare_fn compare;
	union table_parameter parameter;
	lock_t free;
	counter_t counter;
	bool_t inuse;
};

typedef const struct table * const table_t;

typedef union table_ptr {
} tableptr_t;

struct table_op {
	struct table *(*new)(struct table *table);
	void (*dispose)(struct table *table);
	void *(*lookup)(struct table *table, void *element, union table_ptr *ptr, int lookup);
	void *(*get)(struct table *table, union table_ptr *ptr, int get);
	struct table *(*resize)(struct table *table, int size);
};

table_t table_new(int type, int size, int width, compare_fn compare, hash_fn hash_rehash, float occupancy);
void table_dispose(table_t table);

void *table_lookup(table_t table, void *element, tableptr_t *ptr, int lookup);

void *table_get(table_t table, void *element, tableptr_t *ptr, int get);

table_t table_resize(table_t table, int size);

#endif
