#ifndef CHUNK_H
#define CHUNK_H

#include<sys/time.h>

#include"partial.h"

enum chunk_type { CLUSTERED, SECTIONED };

/* orphaned nodes */

struct chunk {
	struct chunk_entry {
		union chunk_base {
			unsigned long clustered;
			struct sectioned_base {
				unsigned section;
				unsigned offset;
			} stretch;
		} base;
		struct chunk_length {
			unsigned total;
			unsigned used;
		} length;
	} entry;
	struct chunk_recent {
		struct chunk *previous, *next;
	} recent;
	struct chunk *left, *right;
};

union chunk_node {
	struct chunk chunk;
	union chunk_node *next;
};

struct chunk_table {
	union chunk_node *base, *avail, *recent;
	int size;
	pthread_mutex_t avail;
};

struct chunk_file_table {
	struct chunk_table memory, data;
};

struct chunk_toc_table {
	struct partial_table disk, core;
	struct partial_table pieces;
};

struct chunk_inode {
	struct partial *name;
	int type;
	short link;
	short owner, group;
	short permission;
	long size;
	struct filetime {
		struct timeval created, accessed, read, written;
	} time;
	union directory_ptr {
		struct chunk_inode *dir, *link;
	} ptr;
	struct partial *toc, *pieces;
	struct chunk_inode *left, *right
};

union chunk_inode_node {
	struct chunk_inode inode;
	union chunk_inode_node *next;
};

struct chunk_inode_table {
	union chunk_inode_node *base, *avail;
	int size;
	pthread_mutex_t avail;
};

struct chunk_fs_table {
	int type;
	void *namebase;
	struct partial_table name;
	struct chunk_file_table file;
	struct chunk_toc_table toc;
	struct chunk_inode_table inode;
};

struct chunk_toc {
	struct partial *whole, *pieces;
};

struct chunk_ptr {
	bool_t pieces;
	short offset;
	struct partial *partial;
};

#define PATH_LEN (256)

struct file;

struct fs_size {
	int block;
	int cache;
	long max, avail;
	struct fs_file {
		int max, count;
	} file;
};

union fs_table {
	struct chunk_fs_table *chunk;
};

struct fs {
	int type;
	struct device *device;
	struct cache *cache;
	char path[PATH_LEN];
	struct file *root;
	struct fs_size size;
	union fs_table table;
};
#define FS_SIZE (sizeof(struct fs))

struct bad {
	long block;
	short count;
};

struct bad_table {
	struct bad *base;
	int size;
};

struct make_arg {
	int type;
	bool_t root;
	char *label;
	struct device *device;
	struct fs_size size;
	struct bad_table bad;
};

enum check_arg { CHECK_FS = 1, RESTORE_FS, TUNE_FS = 4 };

struct mount_arg {
	int type;
	char *device;
	char *dir;
};

struct fs_ops {
	int (*fs_make)(struct make_arg *arg);
	struct fs *(*fs_check)(struct fs *fs, int arg);
	struct fs *(*mount)(struct mount_arg *arg);
	int (*unmount)(struct fs *fs);
	struct fs *(*sync)(struct fs *fs);
};

int chunk_fs_make(struct make_arg *arg) {
	static struct fs *new;

	if(new = malloc(FS_SIZE)) {
		new->type = CHUNK_FS;
	}
}

#endif
