#ifndef CHUNK_H
#define CHUNK_H

#include<sys/time.h>

#include"list.h"
#include"partial.h"

enum chunk_type { CLUSTERED, SECTIONED };

/* orphaned nodes */

struct chunk_entry {
	union chunk_base {
		unsigned long clustered;
		struct sectioned_base {
			unsigned section;
			unsigned offset;
		} stretch;
	} base;
	struct chunk_length {
		unsigned total;
		unsigned used;
	} length;
};

struct chunk {
	struct chunk_entry entry;
	list_ptr_t recent;
};

union chunk_node {
	struct chunk chunk;
	union chunk_node *next;
};

struct chunk_file_table {
	struct chunk *base;
	table_t memory, data;
	list_t least;
};

struct chunk_toc_table {
	struct partial_table disk, core;
	struct partial_table pieces;
};

/* chunk on-disk inode */
struct chunk_inode {
	struct partial *name;
	int type;
	short link;
	short owner, group;
	short permission;
	long size;
	struct file_time {
		struct timeval created, accessed, read, written;
	} time;
	union directory_ptr {
		struct chunk_inode *dir, *link;
	} ptr;
	struct partial *toc, *pieces;
};

union chunk_inode_node {
	struct chunk_inode chunk_inode;
	union chunk_inode_node *next;
};

struct chunk_inode_table {
	table_t table;
	union chunk_inode_node *base, *avail;
	lock_t free;
};

struct chunk_fs_table {
	void *namebase;
	struct partial_table name;
	struct chunk_file_table file;
	struct chunk_toc_table toc;
	struct chunk_inode_table inode;
};

struct chunk_toc_ptr {
	bool_t pieces;
	short offset;
	struct partial *partial;
};

struct chunk_toc {
	struct partial *whole, *pieces;
	struct chunk_toc_ptr ptr;
};

#endif
