#ifndef FILE_H
#define FILE_H

#include<sys/time.h>

#include"fs.h"

#include"chunk.h"

#define NAME_LEN (32)

struct fs;

struct file_info {
	int type;
	short link;
	char name[NAME_LEN];
	short owner, group;
	short permission;
	long length;
	struct file_time {
		struct timeval created, accessed, read, written;
	} time;
};

union file_toc {
	struct chunk_toc chunk;
};

struct file_pos {
	long pos;
	long address;
	int count;
};

/* on-disk inode */
union file_inode {
	struct chunk_inode *chunk;
};

/* in-core inode */
struct inode {
	struct file_info info;
	union file_toc toc;
	struct file_pos pos;
	union file_inode disk;
};

struct file {
	int count;
	struct fs *fs;
	struct file *dir;
	struct inode inode;
	list_t clist;
};

#define ismount(file) ((file)->fs->root == (file))

struct file_ops {
	struct inode *(*file_open)(struct fs *fs, char *relative, int mode);
	long (*file_pos)(struct inode *inode, long *pos);
	long (*file_length)(struct inode *inode, long length);
	struct inode *(*file_sync)(struct inode *inode);
};

#endif
