#ifndef FS_H
#define FS_H

#include<sys/time.h>

#include"device.h"
#include"cache.h"
#include"file.h"

#include"chunk.h"

#define REMOTE_MAX

enum fs_mount { FS_LOCAL = 1, FS_REMOTE };
enum fs_role { FS_LOCAL_SELF = FS_REMOTE_GUEST = 0, FS_REMOTE_HOST = 4 };
enum fs_id { CHUNK };

#define fs_type(mount,role,id) ((((id) * 4) + (role)) * 4) + (mount))
#define fs_id(type) ((type) / 16)
#define fs_role(type)  (((type) / 4) % 4)
#define fs_mount(type) ((type) % 4)

struct file;

struct fs_size {
	int block;
	long max, avail;
	struct fs_file {
		int max, count;
	} file;
};

union fs_table {
	struct chunk_fs_table *chunk;
};

/* On-disk super block */
union fs_super {
	struct chunk_super chunk;
};

/* In-core super block */
struct super {
	struct file *root;
	struct fs_size size;
	union fs_table table;
	union fs_super disk;
};

struct medium {
	struct device *volume;
	struct channel *channel[REMOTE_MOUNT_MAX];
};

struct fs {
	int type;
	struct cache cache;
	struct super super;
	struct medium medium;
};

struct bad {
	long block;
	short count;
};

struct bad_table {
	struct bad *base;
	int size;
};

struct make_arg {
	struct device *volume;
	bool_t root;
	struct fs_size size;
	struct bad_table bad;
};

struct fs_ops {
	int (*make)(struct make_arg *arg);
	struct fs *(*mount)(struct fs *fs);
	int (*unmount)(struct fs *fs);
	struct fs *(*sync)(struct fs *fs);
	struct fs *(*check)(struct fs *fs);
	struct fs *(*tune)(struct fs *fs);
};

#endif
