#ifndef CACHE_H
#define CACHE_H

#include"lock.h"
#include"mutex.h"

#include"file.h"
#include"fs.h"

struct block {
	lock_t lock;
	bool_t write;
	long address;
	void *ptr;
};

struct block_node {
	struct block *block;
	union block_previous {
		struct file *file;
		struct block *previous;
	} previous;
	struct block *next;
};

struct block_pool {
	struct block_node *base, *avail;
	pthread_mutex_t free;
};

struct cache_entry {
	struct block *block; 
	struct cache_entry *left, *right;
	struct cache_entry *previous, *next;
};

struct cache_table {
	struct cache_table_base {
		struct block *block;
		struct cache_entry *cache;
	} base;
	struct cache_entry *root, *head, *least, *avail;
	struct cache_table_mutex {
		pthread_mutex_t root, head, free;
	} mutex;
};

union cache_base {
	void *memory;
	struct file *file;
};

struct cache {
	int size;
	struct fs *fs;
	union cache_base base;
	struct block_pool pool;
	struct cache_table table;
};

void *cache_read(struct cache *cache, long address, void *buffer);
struct cache *cache_write(struct cache *cache, long address, void *buffer);

#endif
