
#include<sys/types.h>

#include"../../include/vm.h"

/* These are arbitrary values, which must be appropriately defined for a specific system */
#define VMS_FRAME_MAX (1073741824)
#define VMS_SPACE_MAX (8192)
#define VMS_LEVEL_MAX (5000)
#define VMS_FAULT_MAX (1024)
#define VMS_INTERVAL_MAX (28416)

/* This is an entry in the address translation table, which will most likely be implemented in some
 * form as h/w. This is a s/w implementation of the table, as a forest of threaded proper trees,
 * which should provide reasonable(near ideal) lookup times, without being too complex or expensive. */
struct map_entry {
	struct entry_flags {
		char valid : 1;
		char modify : 1;
		char reference : 1;
		char protection : 3;
	} flag;
	short memory;
	unsigned page;
	struct entry *left, *right;
};

struct vms_map {
	struct entry base[VMS_FRAME_MAX];
};

struct frame;

struct level {
	struct frame *head;
	struct level *previous, *next;
};

union level_node {
	struct level level;
	union level_node *next;
};

struct frame {
	struct frame_flags {
		char copy : 1;
		char locked : 1;
		char enabled : 1;
		char disabled : 1;
	} flag;
	/* unsigned age; */
	struct frame *previous, *next;
	struct frame_usage {
		struct level *level;
		struct frame *previous, *next;
	} usage;
};
#define FRAME_SIZE (sizeof(struct frame))

struct fault {
	unsigned short complete : 1;
	unsigned short count : 15;
	unsigned page;
	struct fault *left, *right;
};
#define FAULT_SIZE (sizeof(struct fault))

union fault_node {
	struct fault fault;
	union fault_node *next;
};

struct interval {
	unsigned page;
	unsigned short count;
	struct interval *left, *right;
	struct interval *previous, *next;
};

struct vm_root {
	struct entry *map;
	struct fault *fault;
	struct interval *interval;
};

/* Ideally, vm and the associated vms_pool should be implemented as h/w */
struct vm {
	struct file *_virtual;
	unsigned frames;
	struct frame *head;
	struct vm_root root;
};

union vm_node {
	struct vm space;
	union vm_node *next;
};

struct vms_pool {
     union vm_node base[VMS_SPACE_MAX], *avail;
};

struct vms_table {
	struct frame *frame;
	struct usage_table {
		union level_node base[VMS_LEVEL_MAX], *avail;
		struct frame *least;
	} usage;
	struct fault_table {
		union fault_node base[VMS_FAULT_MAX], *avail;
		struct fault *most;
	} fault;
	struct interval_table {
		struct interval base[VMS_INTERVAL_MAX], *head, *least;
	} interval;
};

struct vms_stat;

struct vms {
	/* These fields are part of the h/w */
	void *physical;
	unsigned frames;
	struct vms_map map;
	struct vms_pool pool;
	/* These fields are part of the kernel */
	struct vms_table table;
	struct vms_stat stat;
};

extern struct vms vms;
