#ifndef VMXDEF_H
#define VMXDEF_H

#include<sys/types.h>

#include"../../include/vm.h"

struct vm_frame {
	unsigned page0;
	unsigned char page1 : 3;
	unsigned char dirty : 1;
	unsigned char disable : 1;
	unsigned counter;
	unsigned level;
	unsigned short request, read[2], write[2];
	struct vm_frame *next;
	struct {
	   struct vm_frame *left, *right;
	} lru;
};
#define VM_FRAME_SIZE (sizeof(struct vm_frame))

struct vm_frame_table {
	struct vm_frame *base;
	struct vm_usage_table {
		unsigned char *pos;
		struct vm_usage_level {
			struct vm_frame *used, *unused;
		} *base;
		struct vm_frame *lru;
	} usage;
};
#define VM_FRAME_TABLE_SIZE (sizeof(struct vm_frame_table))

struct vm_fix {
	unsigned count;
	short left, right;
};
#define VM_FIX_SIZE (sizeof(struct vm_fix))

struct vm_fix_table {
	short *base;
	struct vm_association_table {
		struct vm_fix *base;
		short root, lav;
	} association;
};
#define VM_FIX_TABLE_SIZE (sizeof(struct vm_fix_table))

struct vm_request {
	unsigned page0;
	unsigned page1 : 3;
	unsigned frame : 20;
	unsigned count : 8;
	unsigned complete : 1;
	struct vm_request *next;
};
#define VM_REQUEST_SIZE (sizeof(struct vm_request))

struct vm_request_table {
	struct vm_request *base, *avail, **head;
	unsigned char *search;
};
#define VM_REQUEST_TABLE_SIZE (sizeof(struct vm_request_table))

struct vm_table {
	struct vm_fix_table fix[3];
	struct vm_frame_table frame;
	struct vm_request_table request;
};
#define VM_TABLE_SIZE (sizeof(struct vm_table))

struct vm_mutex {
	pthread_mutex_t realloc, sync, load, lru, lav, head, avail;
};
#define VM_MUTEX_SIZE (sizeof(struct vm_mutex))

struct vm_stat;

enum vm_flags { VM_REALLOC = 1, VM_SYNC };

struct vm {
	char flags;
	int _virtual;
	void *physical;
	long long pages, last;
   	int frames;
	short access;
	struct vm_table table;
	struct vm_mutex mutex;
	struct vm_stat stat;
};
#define VM_SIZE (sizeof(struct vm))

#endif
