
#ifndef SHORT_SIZE
#define SHORT_SIZE (sizeof(short))
#endif

#ifndef NODE_MAX
#define NODE_MAX (65535)
#endif

typedef unsigned short node_t;

#ifndef LEVEL_MAX
#define LEVEL_MAX (7)
#endif

typedef struct address {
	node_t node[LEVEL_MAX + 1];
} address_t;

typedef struct edge {
	address_t address;
	struct edge_weight {
		unsigned actual, threshold, max;
	} weight;
} * edge_t;

typedef struct adjacent {
	short count, pos;
	edge_t edge;
} * adjacent;

typedef unsigned char * branch_t;

typedef struct route {
	edge_t edge;
	short length;
	address_t source;
	struct destination {
		bool_t valid;
		node_t local;
		address_t final;
	} destination;
	branch_t branch;
} * route_t;

struct connection {
	short count;
	node_t node, *via;
};

typedef struct chart {
	struct connection connection[LEVEL_MAX + 1];
	void *map[LEVEL_MAX + 1];
} * chart_t;

static inline int nodecmp(node_t *to, node_t *from) {
	return memncmp(&to->rung, &from->rung, (LEVEL_MAX + 1) * SHORT_SIZE);
}

route_t route(route_t route, adjacent_t adjacent, chart_t chart) {
	int above, shift, notfound;
	edge_t farthest, edge, closest;

	if(above = (nodecmp(&route->destination, &adjacent->edge[adjacent->pos].node) > 0)) {
		shift = adjacent->count - adjacent->pos - 1;
		closest = &adjacent->edge[adjacent->count - 1];
	}
	else {
		shift = adjacent->pos + 1;
		closest = adjacent->edge;
	}
	while(shift && (notfound = nodecmp(&route->destination, &closest->node))) {
		shift = (shift / 2) + (shift % 2);
		if(notfound > 0) closest += shift;
		else closest -= shift;
	}
	if(notfound) {
		if(notfound > 0) {
			if(closest < adjacent->edge + adjacent->count - 1 &&
			closer(&route->destination, &closest->node, &(closest + 1)->node)) {
					closest++;
			}
		}
		else {
			if(closest > adjacent->edge &&
			closer(&route->destination, &closest->node, &(closest - 1)->node)) {
					closest--;
			}
		}
		edge = closest;
		farthest = adjacent->edge + adjacent->pos;
		do {
			if(connected(&route->destination, &closest->node, chart)) {
				if(closest->weight.actual <= closest->weight.threshold) {
					edge = closest;
					break;
				}
				if(edge->weight.actual > closest->weight.actual)
					edge = closest;
			}
		} while(above ? --closest > farthest : ++closest < farthest);
	}
	else edge = closest;
	route->edge = edge;
	if(route->length) route->branch[route->length] = edge - adjacent->edge;
	return route;
}
