
#include<signal.h>
#include<pthread.h>
#include<errno.h>

#include"list.h"

#include"lock.h"
#include"deadlock.h"

void lock_init(lock_t *lock) {
	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cancel);
	pthread_setinterruptstate(PTHREAD_INTERRUPT_DISABLE);
	lock->thread = NULL;
	lock->lock = lock->mutex = 0;
	listptr_init(&lock->ptr.lock, LOCK_LIST_TYPE);
	listptr_init(&lock->ptr.deadlock, LOCK_LIST_TYPE);
	pthread_setcancelstate(cancel, NULL);
	pthread_setinterruptstate(PTHREAD_INTERRUPT_ENABLE);
	return;
}

int trylock(lock_t *lock) {
	pthread_t self;

	if(lock->thread == (self = pthread_self())) return 1;
	if(! (lock->thread || lock->lock)) {
		int cancel;

		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cancel);
		pthread_setinterruptstate(PTHREAD_INTERRUPT_DISABLE);
		lock->lock++;
		if(lock->lock == 1) {
			if(lock->mutex || list_insert(self->list.lock, self, &lock->ptr.lock)) {
				lock->thread = self;
				if(! self->process.deadlock_check ||
					list_delete(self->list.deadlock, &lock->ptr.deadlock)) {
					pthread_setcancelstate(cancel, NULL);
					pthread_setinterruptstate(PTHREAD_INTERRUPT_ENABLE);
					return 1;
				}
				pthread_setcancelstate(cancel, NULL);
				pthread_setinterruptstate(PTHREAD_INTERRUPT_ENABLE);
				return 2;
			}
		}
		lock->lock--;
		pthread_setcancelstate(cancel, NULL);
		pthread_setinterruptstate(PTHREAD_INTERRUPT_ENABLE);
		return -1;
	}
	errno = ((lock->thread->process->deadlock_check) && deadlocked(lock)) ? EDEADLK : EBUSY;
	return 0;
}

int lock(lock_t *lock) {
	unsigned retries = 0, counter = 0, start = 0;

	while(trylock(lock) <= 0) {
		if(errno == EDEADLK) {
			if(counter != ++retries) start = retries = counter;
			else {
				if(retries - start >= DEADLOCK_RETRIES) raise(SIG_DEADLK);
			}
		}
		pthread_switch();
		++counter;
	}
	return 1;
}

int unlock(lock_t *lock) {
	pthread_t self;

	if(lock->thread == (self = pthread_self()) && lock->lock == 1) {
		int cancel;

		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cancel);
		pthread_setinterruptstate(PTHREAD_INTERRUPT_DISABLE);
		lock->lock--;
		lock->thread = NULL;
		if(lock->mutex || list_delete(self->list.lock, &lock->ptr.lock)) {
			pthread_setcancelstate(cancel, NULL);
			pthread_setinterruptstate(PTHREAD_INTERRUPT_ENABLE);
			return 1;
		}
		pthread_setcancelstate(cancel, NULL);
		pthread_setinterruptstate(PTHREAD_INTERRUPT_ENABLE);
		return -1;
	}
	return 0;
}
