#ifndef RWLOCK_H
#define RWLOCK_H

#include"../bool.h"
#include"counter.h"

#ifndef EBUSY
#define EBUSY
#endif

#ifndef PTHREAD_RWLOCKATTR_T
#define PTHREAD_RWLOCKATTR_T
typedef int pthread_rwlockattr_t;
#endif

enum rwlock_attribute { RWLOCK_BALANCED, RWLOCK_STRONG_READER, RWLOCK_STRONG_WRITER = 3 };

#ifndef PTHREAD_RWLOCK_T
#define PTHREAD_RWLOCK_T
typedef struct rwlock {
	pthread_rwlockattr_t type;
	counter_t counter;
} pthread_rwlock_t;
#endif

/* These defenitions, structures are part of the user process, and not the kernel process */

/* Maximum number of rwlock's per process */
#ifdef BITWISE
#define RWLOCK_MAX ((PTHREADS_MAX) << (2))
#else
#define RWLOCK_MAX ((PTHREADS_MAX) * (4))
#endif

struct rwlock_entry {
	bool_t write;
	pthread_rwlock_t *rwlock;
	struct rwlock_entry *next;
};

union rwlock_node {
	struct rwlock_entry rwlock;
	union rwlock_node *next;
};

struct rwlock_pool {
	union rwlock_node base[RWLOCK_MAX], *avail;
	lock_t free;
};

int pthread_rwlock_init(pthread_rwlock_t *rwlock, pthread_rwlockattr_t *attr);
int pthread_rwlock_destroy(pthread_rwlock_t *rwlock);

/* This function is non-standard , use macros instead */
int pthread_rwlock_trylock(pthread_rwlock_t *rwlock, bool_t write);
#define pthread_rwlock_tryrdlock(rwlock) (! pthread_rwlock_trylock(rwlock, FALSE))
#define pthread_rwlock_trywrlock(rwlock) (! pthread_rwlock_trylock(rwlock, TRUE))

/* This function is non-standard , use macros instead */
int pthread_rwlock_lock(pthread_rwlock_t *rwlock, bool_t write);
#define pthread_rwlock_rdlock(rwlock) (! pthread_rwlock_trylock(rwlock, FALSE))
#define pthread_rwlock_wrlock(rwlock) (! pthread_rwlock_trylock(rwlock, TRUE))

int pthread_rwlock_unlock(pthread_rwlock_t *rwlock);

#endif
