
#include<errno.h>

#include"rwlock.h"

int pthread_rwlock_init(pthread_rwlock_t *rwlock, pthread_rwlockattr_t *attr) {
	if(rwlock->dead || *attr / PTHREAD_INITIALIZE) {
		rwlock->type = *attr % PTHREAD_INITIALIZE;
		rwlock->counter.balanced.up = 0;
		do {
			rwlock->dead = FALSE;
		}
		 counter_init(&rwlock->counter, rwlock->type);
	}
	return 0;
}

int pthread_rwlock_destroy(pthread_rwlock_t *rwlock) {
	if(counter_active(&rwlock->counter)) errno = EACTIVE;
       	return 0;
}

static struct rwlock_entry *find_rwlock(pthread_rwlock_t *rwlock, list_ptr_t *ptr) {
	struct rwlock_entry *request;
	pthread_t self;

	self = pthread_self();
	request = list_get(self->list.rwlock, GET_FIRST, ptr);
	while(request && request->rwlock != rwlock) {
		request = list_get(self->list.rwlock, GET_NEXT, ptr);
	}
	return request;
}

static int add_rwlock(pthread_rwlock_t *rwlock, bool_t write) {
	union rwlock_node *avail;
	pthread_t self;

	self = pthread_self();
	while(! (trylock(&rwlock.free) && avail = rwlock_pool.avail)) {
		unlock(&rwlock.free);
		pthread_switch();
	}
	rwlock_pool.avail = avail->next;
	unlock(&rwlock_pool.free);
	avail->rwlock.write = write;
	avail->rwlock.rwlock = rwlock;
	if(! list_insert(self->list.rwlock, &avail->rwlock, NULL)) {
		lock(&rwlock_pool.free);
		avail->next = rwlock_pool.avail;
		rwlock_pool.avail = avail;
		unlock(&rwlock_pool.free);
		return 0;
	}
	return 1;
}

int pthread_rwlock_trylock(pthread_rwlock_t *rwlock, bool_t write) {
	int success, type;

	type = rwlock->type + (write ? SEQUENTIAL : PARALLEL);
	if (success = counter_tryup(&rwlock->counter, type) == 1) {
		if(! find_rwlock(rwlock, NULL)) {
			if(! add_rwlock(rwlock, write)) success = 3;
		}
	}
	if(success > 1) {
		counter_down(&rwlock->counter, type);
		if(success != 3) errno = EBUSY;
	}
	return success;
}

int pthread_rwlock_timedlock(pthread_rwlock_t *rwlock, bool_t write, struct timespec *abstime) {
	struct rwlock_entry *entry;

	if(entry = find_rwlock(rwlock, NULL)) {
		if(write || entry->write) {
			errno = EBUSY;
			return 0;
		}
	}
	else {
		int type;

		type = rwlock->type + (write ? SEQUENTIAL : PARALLEL);
		if(counter_timedup(&rwlock->counter, type, abstime)) {
			if(! add_rwlock(rwlock, write)) {
				counter_down(&rwlock->counter, type);
				return 0;
			}
		}
		else return 0;
	}
	return 1;
}

int pthread_rwlock_unlock(pthread_rwlock_t *rwlock) {
	struct rwlock_entry *entry;
	union rwlock_node *avail;
	list_ptr_t ptr;
	pthread_t self;
	int type;

	while((entry = find_rwlock(rwlock, &ptr)) &&
		(! list_delete(self->list.rwlock, &ptr)));
	if(entry) {
		self = pthread_self();
		type = entry->rwlock->type + (entry->write ? SEQUENTIAL : PARALLEL);
		counter_down(&entry->rwlock->counter, type);
		lock(&rwlock_pool.free);
		avail->next = rwlock_pool.avail;
		rwlock_pool.avail = avail;
		unlock(&rwlock_pool.free);
		return 0;
	}
	return 1;
}
