
#include<errno.h>

#include"rwlock.h"

int pthread_rwlock_init(pthread_rwlock_t *rwlock, pthread_rwlockattr_t *attr) {
	rwlock->counter.rwlock = 1;
	if(rwlock->dead && ! (rwlock->once || counter_active(&rwlock->counter))) {
		rwlock->type = *attr;
		rwlock->dead = FALSE;
		while(! rwlock->once) {
			rwlock->once++;
			if(rwlock->once == 1) {
		 		counter_init(&rwlock->counter, rwlock->type);
				break;
			}
			rwlock->once--;
		}
		return 0;
	}
	return 1;
}

int pthread_rwlock_destroy(pthread_rwlock_t *rwlock) {
	rwlock->once = 0;
	rwlock->dead = TRUE;
	if(counter_active(&rwlock->counter)) {
		rwlock->dead = FALSE;
		errno = EACTIVE;
		return 1;
	}
       	return 0;
}

static struct rwlock_entry *find_rwlock(pthread_rwlock_t *rwlock, bool_t write, listptr_t *ptr) {
	struct rwlock_entry *request;
	pthread_t self;

	self = pthread_self();
	request = list_get(self->list.rwlock, GET_FIRST, ptr);
	while(request && (request->rwlock != rwlock || rwlock->write < write)) {
		request = list_get(self->list.rwlock, GET_NEXT, ptr);
	}
	return request;
}

static int add_rwlock(pthread_rwlock_t *rwlock, bool_t write) {
	union rwlock_node *avail;
	pthread_t self;

	self = pthread_self();
	lock(&rwlock_pool.free);
	if(avail = rwlock_pool.avail) {
		rwlock_pool.avail = avail->next;
		unlock(&rwlock_pool.free);
		avail->rwlock.write = write;
		avail->rwlock.rwlock = rwlock;
		if(list_insert(self->list.rwlock, &avail->rwlock, NULL)) return 1;
		lock(&rwlock_pool.free);
		avail->next = rwlock_pool.avail;
		rwlock_pool.avail = avail;
	}
	unlock(&rwlock_pool.free);
	return 0;
}

int pthread_rwlock_trylock(pthread_rwlock_t *rwlock, bool_t write) {
	int success, op;
	struct rwlock_entry *entry;

	op = (write ? SEQUENTIAL : PARALLEL);
	if ((success = counter_tryup(&rwlock->counter, op)) == 1) {
		if(! entry = find_rwlock(rwlock, write, NULL)) {
			if(! add_rwlock(rwlock, write)) success = -1;
		}
		else {
			if(write && ! entry->write) {
				success = -2;
				errno = EBUSY;
			}
			else success = 3;
		}
	}
	if(success < 0 || success > 1) counter_down(&rwlock->counter, op);
	return success;
}

int pthread_rwlock_timedlock(pthread_rwlock_t *rwlock, bool_t write, struct timespec *abstime) {
	struct rwlock_entry *entry;

	if(! entry = find_rwlock(rwlock, write, NULL)) {
		int op;

		op = (write ? SEQUENTIAL : PARALLEL);
		if(counter_timedup(&rwlock->counter, op, abstime) > 0) {
			if(! add_rwlock(rwlock, write)) {
				counter_down(&rwlock->counter, op);
				return -1;
			}
			return 0;
		}
	}
	return 1;
}

int pthread_rwlock_unlock(pthread_rwlock_t *rwlock) {
	struct rwlock_entry *entry;
	union rwlock_node *avail;
	listptr_t ptr;
	pthread_t self;
	int op;

	while((entry = find_rwlock(rwlock, FALSE, &ptr)) &&
		(! list_delete(self->list.rwlock, &ptr)));
	if(entry) {
		self = pthread_self();
		op = (entry->write ? SEQUENTIAL : PARALLEL);
		counter_down(&entry->rwlock->counter, op);
		lock(&rwlock_pool.free);
		avail->next = rwlock_pool.avail;
		rwlock_pool.avail = avail;
		unlock(&rwlock_pool.free);
		return 0;
	}
	return 1;
}
